/*
 * Class for network communication between game server and zone master.
 * 
 * @author Frode Voll Aasen
 * @author Tom Christian Bjrlo Johannesen
 * 
 */

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintStream;
import java.net.Socket;
import java.util.ArrayList;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.InputSource;

import com.sun.org.apache.xerces.internal.parsers.DOMParser;


public class ZoneMasterConnection extends Thread {
	private GameServer gs;
	private Socket zm;
	private ArrayList<SDLTimer> timers = new ArrayList<SDLTimer>(10);
	private ArrayList<Signal> input = new ArrayList<Signal>(100);
	/*
	 * Class constructor
	 * 
	 * @param gs	Reference to GameServer
	 * @param zm	Connection to zone master
	 */
	public ZoneMasterConnection(GameServer gs, Socket zm) {
		this.gs = gs;
		this.zm = zm;
	}
	
	/*
	 * Retrieves the IP address of the zone master
	 * 
	 * @return Zone master IP
	 */
	public String getZoneMasterIP() {
		return zm.getInetAddress().getHostAddress();
	}
	
	/*
	 * Retrieves the port of the zone master connection
	 * 
	 * @return Zone master connection port
	 */
	public int getZoneMasterPort() {
		return zm.getPort();
	}
	
	/*
	 * Class for handling the reading and writing over the network
	 * 
	 */
	private class socket extends Thread {
		private BufferedReader network;
		public boolean running = true;
		private ZoneMasterConnection parent;
		private Socket zm;
		
		public socket(Socket zm, ZoneMasterConnection parent) {
			try {
				network = new BufferedReader(new InputStreamReader(zm.getInputStream()));
				
			} catch (IOException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			this.parent = parent;
			this.zm = zm;
		}
		
		public void run() {
			while(running) {
				try {
					//if(network.ready()) {
						String network_in;
						network_in = network.readLine();
						network_in.trim();
						if(!network_in.equals("")) {
							String[] signal_in = network_in.split(";");
							ArrayList netInParam = new ArrayList();
							for(int i = 1; i < signal_in.length; i++) {
								netInParam.add(signal_in[i].trim());
							}
							parent.addInput(new Signal(signal_in[0].trim(), netInParam, false));
						}
					//}
				} catch (IOException e) {
					// TODO Auto-generated catch block
					running = false;
					try {
						zm.close();
					} catch (IOException e2) {
						
					}
					System.err.println("ZoneMaster disconnected from GS");
					parent.addInput(new Signal("Disconnect", null, true));
				}
			}
		} 
	}
	
	public void run() {
		System.out.println("ZoneMasterConnection starting... Connection from " + zm.getRemoteSocketAddress().toString());
		String state = "Idle";
		boolean running = true;
		boolean isEmpty = true;
		
		socket networkthread = this.new socket(zm, this);
		networkthread.start();
		
		try {
			while(running) {
				while(isEmpty) { // While input queue has no input
					for(int i = 0; i < timers.size(); i++) {
						if(timers.get(i).isExpired()) {
							this.addInput(timers.remove(i).getSignal()); //TODO should be added with priority
						}
					}
					synchronized(input) { // protect variable
						if(!input.isEmpty()) isEmpty = false;
					}
					Thread.sleep(5);
				}
				isEmpty = true;
				synchronized(input) {
					Signal s = input.remove(0);
					String signal = s.name;
					System.out.println("ZoneMasterConnection received signal: " + signal + " in state " + state);
					if(state.equals("Idle")) {
						if(signal.equals("ZMConnectGS")) {
							s.parameters.add(this);
							gs.zm.addInput(new Signal("GetZoneInfo", s.parameters, false));
							state = "Waiting";
						}
					}
					else if(state.equals("Waiting")) {
						if(signal.equals("ZoneInfo")) {
							this.signalZM(s);
							state = "Connected";
						}
					}
					else if(state.equals("Connected")) {
						if(signal.equals("ObjectAdd")) {
							this.signalZM(s);
						}
						if(signal.equals("ChallengeAnswer")) {
							this.signalZM(s);
						}
						if(signal.equals("NewAdjacentZoneMaster")) {
							this.signalZM(s);
						}
						if(signal.equals("StopManagingZone")) {
							this.signalZM(s);
						}
						if(signal.equals("ClientHasNewMaster")) {
							GameObjectList gol = parseXMLString((String)s.parameters.get(0));
							PlayerCharacter pc = (PlayerCharacter) gol.GameObjects.get(0);
							ArrayList param = new ArrayList();
							param.add(pc);
							s.parameters = param;
							gs.zm.addInput(s);
						}
						if(signal.equals("Disconnect")) {
							running = false;
						}
						if(signal.equals("ClientTimeout")) {
							gs.zm.addInput(s);
						}
						if(signal.equals("Backup")) {
							gs.zm.addInput(s);
						}
						if(signal.equals("ClientConnected")) {
							gs.zm.addInput(s);
						}
						if(signal.equals("ClientDisconnect")) {
							this.signalZM(s);
						}
					}
				}
			}
		} catch (InterruptedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	/*
	 * Adds a signal to the back of the signal queue.
	 * 
	 * @param input The signal to be added
	 */
	public void addInput(Signal input) {
		synchronized(this.input) {
			this.input.add(input);
		}
	}
	
	/*
	 * Sends a signal over the network
	 * 
	 * @param output The signal to be sent
	 */
	public void signalZM(Signal output) {
		try {
			PrintStream os = new PrintStream(zm.getOutputStream());
			String out = output.name;
			if(output.parameters != null)
				for(int i = 0; i < output.parameters.size(); i++) {
					out = out + ";" + output.parameters.get(i).toString();
				}
			os.print(out + "\n");
		} catch (IOException e) {
			System.err.println(e);
		}
	}
	
	/*
	 * Parses an XML string to a list of GameObjects
	 * 
	 * @param xml	XML string
	 * @return		GameObjectList containing the objects
	 */
	public GameObjectList parseXMLString(String xml) {
		GameObjectList gol = new GameObjectList();
		
		try {
			DOMParser parser = new DOMParser();
			parser.parse(new InputSource(new java.io.StringReader(xml)));
			Document doc = parser.getDocument();

			doc.getDocumentElement().normalize();
			
			NodeList nodeLst = doc.getElementsByTagName("GameObject");
			for (int s = 0; s < nodeLst.getLength(); s++) {
				Node fstNode = nodeLst.item(s);
				if (fstNode.getNodeType() == Node.ELEMENT_NODE) {			  
					Element fstElmnt = (Element) fstNode;
					
					NodeList oidElmntLst = fstElmnt.getElementsByTagName("objectID");
					Element oidElmnt = (Element) oidElmntLst.item(0);
					NodeList oid = oidElmnt.getChildNodes();
					int oid_ = Integer.parseInt(((Node) oid.item(0)).getNodeValue());
					
					NodeList typeElmntLst = fstElmnt.getElementsByTagName("type");
					Element typeElmnt = (Element) typeElmntLst.item(0);
					NodeList type = typeElmnt.getChildNodes();
					String type_ = ((Node) type.item(0)).getNodeValue();
					
					NodeList posXElmntLst = fstElmnt.getElementsByTagName("posX");
					Element posXElmnt = (Element) posXElmntLst.item(0);
					NodeList posX = posXElmnt.getChildNodes();
					double posX_ = Double.parseDouble(((Node) posX.item(0)).getNodeValue());
					
					NodeList posYElmntLst = fstElmnt.getElementsByTagName("posY");
					Element posYElmnt = (Element) posYElmntLst.item(0);
					NodeList posY = posYElmnt.getChildNodes();
					double posY_ = Double.parseDouble(((Node) posY.item(0)).getNodeValue());
					
					NodeList posZElmntLst = fstElmnt.getElementsByTagName("posZ");
					Element posZElmnt = (Element) posZElmntLst.item(0);
					NodeList posZ = posZElmnt.getChildNodes();
					double posZ_ = Double.parseDouble(((Node) posZ.item(0)).getNodeValue());
					
					NodeList rotXElmntLst = fstElmnt.getElementsByTagName("rotX");
					Element rotXElmnt = (Element) rotXElmntLst.item(0);
					NodeList rotX = rotXElmnt.getChildNodes();
					double rotX_ = Double.parseDouble(((Node) rotX.item(0)).getNodeValue());
					
					NodeList rotYElmntLst = fstElmnt.getElementsByTagName("rotY");
					Element rotYElmnt = (Element) rotYElmntLst.item(0);
					NodeList rotY = rotYElmnt.getChildNodes();
					double rotY_ = Double.parseDouble(((Node) rotY.item(0)).getNodeValue());
					
					NodeList rotZElmntLst = fstElmnt.getElementsByTagName("rotZ");
					Element rotZElmnt = (Element) rotZElmntLst.item(0);
					NodeList rotZ = rotZElmnt.getChildNodes();
					double rotZ_ = Double.parseDouble(((Node) rotZ.item(0)).getNodeValue());
					
					GameObject go = new GameObject(oid_, type_, posX_, posY_, posZ_, rotX_, rotY_, rotZ_);
					gol.add(go);
				}
			}
			
			nodeLst = doc.getElementsByTagName("PlayerCharacter");
			for (int s = 0; s < nodeLst.getLength(); s++) {
				Node fstNode = nodeLst.item(s);
				if (fstNode.getNodeType() == Node.ELEMENT_NODE) {			  
					Element fstElmnt = (Element) fstNode;
					
					NodeList oidElmntLst = fstElmnt.getElementsByTagName("objectID");
					Element oidElmnt = (Element) oidElmntLst.item(0);
					NodeList oid = oidElmnt.getChildNodes();
					int oid_ = Integer.parseInt(((Node) oid.item(0)).getNodeValue());
					
					NodeList posXElmntLst = fstElmnt.getElementsByTagName("posX");
					Element posXElmnt = (Element) posXElmntLst.item(0);
					NodeList posX = posXElmnt.getChildNodes();
					double posX_ = Double.parseDouble(((Node) posX.item(0)).getNodeValue());
					
					NodeList posYElmntLst = fstElmnt.getElementsByTagName("posY");
					Element posYElmnt = (Element) posYElmntLst.item(0);
					NodeList posY = posYElmnt.getChildNodes();
					double posY_ = Double.parseDouble(((Node) posY.item(0)).getNodeValue());
					
					NodeList posZElmntLst = fstElmnt.getElementsByTagName("posZ");
					Element posZElmnt = (Element) posZElmntLst.item(0);
					NodeList posZ = posZElmnt.getChildNodes();
					double posZ_ = Double.parseDouble(((Node) posZ.item(0)).getNodeValue());
					
					NodeList rotXElmntLst = fstElmnt.getElementsByTagName("rotX");
					Element rotXElmnt = (Element) rotXElmntLst.item(0);
					NodeList rotX = rotXElmnt.getChildNodes();
					double rotX_ = Double.parseDouble(((Node) rotX.item(0)).getNodeValue());
					
					NodeList rotYElmntLst = fstElmnt.getElementsByTagName("rotY");
					Element rotYElmnt = (Element) rotYElmntLst.item(0);
					NodeList rotY = rotYElmnt.getChildNodes();
					double rotY_ = Double.parseDouble(((Node) rotY.item(0)).getNodeValue());
					
					NodeList rotZElmntLst = fstElmnt.getElementsByTagName("rotZ");
					Element rotZElmnt = (Element) rotZElmntLst.item(0);
					NodeList rotZ = rotZElmnt.getChildNodes();
					double rotZ_ = Double.parseDouble(((Node) rotZ.item(0)).getNodeValue());
					
					NodeList nameElmntLst = fstElmnt.getElementsByTagName("name");
					Element nameElmnt = (Element) nameElmntLst.item(0);
					NodeList name = nameElmnt.getChildNodes();
					String name_ = ((Node) name.item(0)).getNodeValue();
					
					NodeList movingElmntLst = fstElmnt.getElementsByTagName("moving");
					Element movingElmnt = (Element) movingElmntLst.item(0);
					NodeList moving = movingElmnt.getChildNodes();
					boolean moving_ = Boolean.parseBoolean(((Node) moving.item(0)).getNodeValue());
					
					PlayerCharacter go = new PlayerCharacter(oid_, posX_, posY_, posZ_, rotX_, rotY_, rotZ_, name_);
					go.moving = moving_;
					gol.add(go);
				}
			}
			
		} catch (Exception e) {
			e.printStackTrace();
		}
		return gol;
	}
}

